//this file is part of notepad++
//Copyright (C)2003 Don HO ( donho@altern.org )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "PluginInterface.h"
#include <list>
#include <string>

const char PLUGIN_NAME[] = "Multi-Clipboard";
const int nbFunc = 1;

NppData nppData;
FuncItem funcItem[nbFunc];


// ID for cut and copy function from npp's resource.h
#define IDM_EDIT_CUT  42001
#define IDM_EDIT_COPY 42002
// ID for popup menu item
#define MULTI_COPY_MENU_CMD 52000

// Global variables
std::list<std::string> copyTextList;	// The list of text that is copied
const int MAX_LIST_SIZE = 10;			// Max number of items in the list
const int MENU_TEXT_LENGTH = 42;		// Number of char to display in the popup menu
WNDPROC oldNppWndProc = 0;				// nppData._nppHandle's Windows procedure
WNDPROC oldScintillaMainWndProc = 0;	// nppData._scintillaMainHandle's Windows procedure
WNDPROC oldScintilla2ndWndProc = 0;		// nppData._scintillaSecondHandle's Windows procedure
HMENU hPasteMenu = 0;					// Handle to popup menu
bool useMouseCoords = false;			// Use mouse coords to display popup menu?


// function prototypes
void multiClipboardPaste();
void createTextListMenu();
LRESULT CALLBACK multiClipboardSubClassWndProc(HWND, UINT, WPARAM, LPARAM);
void onCopyText();
void createMenuText( std::string str, char * menuTextBuf );
void cleanup();


BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  reasonForCall,
                       LPVOID lpReserved )
{
    switch (reasonForCall)
    {
    case DLL_PROCESS_ATTACH:
        funcItem[0]._pFunc = multiClipboardPaste;
        strcpy(funcItem[0]._itemName, "Paste");
        funcItem[0]._init2Check = false;
        // Shortcut :
        // Following code makes the first command
        // bind to the shortcut Ctrl-Shift-V
        funcItem[0]._pShKey = new ShortcutKey;
        funcItem[0]._pShKey->_isAlt = false;
        funcItem[0]._pShKey->_isCtrl = true;
        funcItem[0]._pShKey->_isShift = true;
        funcItem[0]._pShKey->_key = 0x56; //VK_V
        break;

    case DLL_PROCESS_DETACH:
        // Don't forget to deallocate your shortcut here
        delete funcItem[0]._pShKey;
        cleanup();
        break;

    case DLL_THREAD_ATTACH:
        break;

    case DLL_THREAD_DETACH:
        break;
    }

    return TRUE;
}

extern "C" __declspec(dllexport) void setInfo(NppData notpadPlusData)
{
    nppData = notpadPlusData;

    oldNppWndProc = (WNDPROC) SetWindowLong( nppData._nppHandle, GWL_WNDPROC, (LONG) multiClipboardSubClassWndProc );
    oldScintillaMainWndProc = (WNDPROC) SetWindowLong( nppData._scintillaMainHandle, GWL_WNDPROC, (LONG) multiClipboardSubClassWndProc );
    oldScintilla2ndWndProc = (WNDPROC) SetWindowLong( nppData._scintillaSecondHandle, GWL_WNDPROC, (LONG) multiClipboardSubClassWndProc );
}

extern "C" __declspec(dllexport) const char * getName()
{
    return PLUGIN_NAME;
}

extern "C" __declspec(dllexport) FuncItem * getFuncsArray(int *nbF)
{
    *nbF = nbFunc;
    return funcItem;
}

HWND getCurrentHScintilla(int which)
{
    return (which == 0)?nppData._scintillaMainHandle:nppData._scintillaSecondHandle;
};

extern "C" __declspec(dllexport) void beNotified(SCNotification *notifyCode)
{/*
    switch (notifyCode->nmhdr.code)
    {
        case SCN_CHARADDED:
        {
        }
        break;
    }*/
}

// Here you can process the Npp Messages
// I will make the messages accessible little by little, according to the need of plugin development.
// Please let me know if you need to access to some messages :
// http://sourceforge.net/forum/forum.php?forum_id=482781
//
extern "C" __declspec(dllexport) LRESULT messageProc(UINT Message, WPARAM wParam, LPARAM lParam)
{
    return TRUE;
}


void multiClipboardPaste()
{
    // Get current scintilla editor
    int currentEdit;
    ::SendMessage(nppData._nppHandle, WM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
    HWND hCurrScintilla = getCurrentHScintilla(currentEdit);

	POINT pt;	// Point to display pop-up menu
	if ( useMouseCoords )
	{
		::GetCursorPos( &pt );	// Mouse cursor position in screen coords
	}
	else
	{
		// Get current cursor position
		int currentPos = ::SendMessage( hCurrScintilla, SCI_GETCURRENTPOS, 0, 0 );
		pt.x = ::SendMessage( hCurrScintilla, SCI_POINTXFROMPOSITION, 0, (LPARAM)currentPos );
		pt.y = ::SendMessage( hCurrScintilla, SCI_POINTYFROMPOSITION, 0, (LPARAM)currentPos );
		ClientToScreen( hCurrScintilla, &pt );
	}

    // Popup the menu
    int id = ::TrackPopupMenu( hPasteMenu, TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RETURNCMD | TPM_LEFTBUTTON,
        pt.x, pt.y, 0, nppData._nppHandle, 0 );

    // Check menu selection
    if ( id >= MULTI_COPY_MENU_CMD && id < (MULTI_COPY_MENU_CMD + MAX_LIST_SIZE) )
    {
        // Get selected item
        int idx = id - MULTI_COPY_MENU_CMD;
        std::list< std::string >::iterator iter = copyTextList.begin();
        for ( int i = 0; i < idx; i++, iter++ );	// Move the iterator to the correct text
        std::string pasteText = *iter;

        // Paste the text into the editor
        ::SendMessage( hCurrScintilla, SCI_REPLACESEL, 0, (LPARAM) pasteText.c_str() );

        // Move the selected item to the front of the list
        if ( idx > 0 )
        {
            // Set selection item to front of list
            copyTextList.erase( iter );
            copyTextList.push_front( pasteText );

            // Copy selected text to clipboard
            ::SendMessage( hCurrScintilla, SCI_COPYTEXT, (WPARAM)pasteText.length(), (LPARAM) pasteText.c_str() );

            // Recreate the menu
            createTextListMenu();
        }
    }
}


// The subclassed windows procedure for npp
LRESULT CALLBACK multiClipboardSubClassWndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
    if ( msg == WM_COMMAND )
    {
        int id = LOWORD(wp);
        if ( id == IDM_EDIT_COPY || id == IDM_EDIT_CUT )
        {
            onCopyText();
        }
    }
    else if ( msg == WM_KEYDOWN )
    {
        if ( wp == 0x43 || wp == 0x58 )	// C or X key
        {
            SHORT keyState = ::GetAsyncKeyState( VK_CONTROL );
            if ( keyState & 0x8000 )
            {
                onCopyText();
            }
        }
    }
	else if ( msg == WM_MBUTTONDOWN )
	{
		if ( copyTextList.size() > 0 )	// Make sure there is some text to paste
		{
			// Get current scintilla editor
			int currentEdit;
			::SendMessage(nppData._nppHandle, WM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
			HWND hCurrScintilla = getCurrentHScintilla(currentEdit);

			// Set cursor position to mouse position
			int currPos = ::SendMessage( hCurrScintilla, SCI_POSITIONFROMPOINT, LOWORD(lp), HIWORD(lp) );
			::SendMessage( hCurrScintilla, SCI_GOTOPOS, currPos, 0 );

			if ( MK_SHIFT & wp )	// if Shift key is pressed
			{
				useMouseCoords = true;	// Display menu at the mouse position

				// Pop up multipaste menu
				multiClipboardPaste();

				useMouseCoords = false;	// Unset this so that next time menu can be displayed at the right position
			}
			else
			{
				// Do normal paste
				// Paste the text into the editor
				std::list< std::string >::iterator iter = copyTextList.begin();
				::SendMessage( hCurrScintilla, SCI_REPLACESEL, 0, (LPARAM) iter->c_str() );
			}
		}
	}

    // Let the original windows procedure process the messages
    // Find the correct windows procedure to call
    if ( nppData._scintillaMainHandle == hwnd )	// Main Scintilla Window
    {
        return CallWindowProc( oldScintillaMainWndProc, hwnd, msg, wp, lp );
    }
    if ( nppData._scintillaSecondHandle == hwnd )	// Second Scintilla Window
    {
        return CallWindowProc( oldScintilla2ndWndProc, hwnd, msg, wp, lp );
    }
    // Notepad++ Window
    return CallWindowProc( oldNppWndProc, hwnd, msg, wp, lp );
}


void onCopyText()
{
    // Get current scintilla editor
    int currentEdit;
    ::SendMessage(nppData._nppHandle, WM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
    HWND hCurrScintilla = getCurrentHScintilla(currentEdit);

    // Get length of selected text
    int textLength = ::SendMessage( hCurrScintilla, SCI_GETSELTEXT, 0, 0 );
    if ( 0 == textLength )
    {
        // No text selected, return
        return;
    }

    // Allocate temporary memory for text
    char * pTempText = new char[textLength];
    // Get selected text
    ::SendMessage( hCurrScintilla, SCI_GETSELTEXT, 0, (LPARAM)pTempText );

    // Find if the text already exist in the list
    for ( std::list<std::string>::iterator iter = copyTextList.begin();
        iter != copyTextList.end(); iter++ )
    {
        if ( *iter == pTempText )
        {
            // Text already exist, remove it from the list. It will be added to the front further down
            copyTextList.erase( iter );
            break;
        }
    }
    // Add text to the front of the list
    copyTextList.push_front( std::string( pTempText ) );
    while ( copyTextList.size() > MAX_LIST_SIZE )
    {
        // Trim list from the back until the size is correct
        copyTextList.pop_back();
    }

    // Delete temporary memory
    delete [] pTempText;

    // Recreate the menu
    createTextListMenu();
}


// Creates a menu that contains the list of the text
// Returns NULL if unsuccessful
void createTextListMenu()
{
    if ( hPasteMenu )
    {
        ::DestroyMenu( hPasteMenu );
        hPasteMenu = 0;
    }

    hPasteMenu = CreatePopupMenu();
    if ( NULL == hPasteMenu )
    {
        return;
    }

    char menuText[MENU_TEXT_LENGTH] = "";

    // Loop through the list and append every one as a menu item
    std::list<std::string>::iterator iter;
    int idx = 0;
    for ( iter = copyTextList.begin(); iter != copyTextList.end(); iter++, idx++ )
    {
        // Create a menu ID based on the its location in the list
        unsigned int menuId = MULTI_COPY_MENU_CMD + idx;
        createMenuText( *iter, menuText );
        BOOL result = AppendMenu( hPasteMenu, MF_STRING, menuId, menuText );
        if ( !result )
        {
            // In case of error, does nothing as yet
            continue;
        }
    }
}


// This function creates the text that will be displayed in the popup menu based on the copied text
void createMenuText( std::string str, char * menuTextBuf )
{
    unsigned int bufIndex = 0;
    bool addedAmpersand = false;
    for ( unsigned int i = 0; i < str.length() && bufIndex < (MENU_TEXT_LENGTH-1); i++, bufIndex++ )
    {
        char c = str[i];

        // Build up the menu text char by char
        if ( c == '\r' || c == '\n' || c == '\t' )	// Whitespace char
        {
            menuTextBuf[bufIndex] = ' ';	// Substitute with a blank space
        }
        else if ( (c >= 'A' && c <= 'Z') ||
                  (c >= 'a' && c <= 'z') ||
                  (c >= '0' && c <= '9') )	// Printable characters
        {
            if ( !addedAmpersand )	// If its the first printable char in the menu text, ...
            {
                menuTextBuf[bufIndex] = '&';	// Add '&' as menu shortcut
                bufIndex++;
                addedAmpersand = true;
            }
            menuTextBuf[bufIndex] = str[i];		// Add char
        }
        else
        {
            menuTextBuf[bufIndex] = str[i];		// Add char
        }
    }
    menuTextBuf[bufIndex] = '\0';	// Add null terminator
}


void cleanup()
{
    // Destroy the popup menu
    if ( 0 != hPasteMenu )
    {
        ::DestroyMenu( hPasteMenu );
        hPasteMenu = 0;
    }

    // Unset the windows subclassing
    SetWindowLong( nppData._nppHandle, GWL_WNDPROC, (LONG)oldNppWndProc );
    SetWindowLong( nppData._scintillaMainHandle, GWL_WNDPROC, (LONG)oldScintillaMainWndProc );
    SetWindowLong( nppData._scintillaSecondHandle, GWL_WNDPROC, (LONG)oldScintilla2ndWndProc );
}